﻿Imports RMCLinkNET

''' <summary>
''' A demo application demonstrating how to use RMCLink with the Visual Basic
''' programming language to interface with a Delta RMC.
''' </summary>
Public Class MonitorMain

    ''' <summary>
    ''' The ID of the last plot uploaded from the RMC.
    ''' When the plot ID read from the RMC differs from this one, we know that
    ''' a new plot has been generated, and can download it.
    ''' </summary>
    Dim LastPlotID As Integer = 0

    ''' <summary>
    ''' The object used for interfacing with the RMC.
    ''' </summary>
    Private RMC As RMCLink

    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        ' Since we are initially disconnected from the RMC, enable and disable buttons and labels accordingly.
        ' TODO: move code setting UI to 'disconnected' state into its own function
        ConnectButton.Enabled = True
        DisconnectButton.Enabled = False

        ' Initialize default plot items.
        If My.Settings.PlotDataItems Is Nothing Then
            My.Settings.PlotDataItems = New Specialized.StringCollection
            My.Settings.PlotDataItems.Add("Targets Pos")
            My.Settings.PlotDataItems.Add("pu")
            My.Settings.PlotDataItems.Add("Actual Pos")
            My.Settings.PlotDataItems.Add("pu")
            My.Settings.PlotDataItems.Add("Target Vel")
            My.Settings.PlotDataItems.Add("pu/s")
            My.Settings.PlotDataItems.Add("Actual Vel")
            My.Settings.PlotDataItems.Add("pu/s")
            My.Settings.PlotDataItems.Add("Ctrl Output")
            My.Settings.PlotDataItems.Add("%")
        End If

        ' Set up the columns for the plot quantity item list.
        Dim dt As DataTable = New Data.DataTable("ParentTable")
        dt.Columns.Add("Item Name", GetType(String))
        dt.Columns.Add("Units", GetType(String))
        Me.dgPlotItems.DataSource = dt


        ' Restore saved plot quantities from settings.
        Dim i As Integer
        For i = 0 To My.Settings.PlotDataItems.Count - 1
            dt.Rows.Add(New String() {My.Settings.PlotDataItems.Item(i), My.Settings.PlotDataItems.Item(i + 1)})
            i += 1
        Next
    End Sub

    Private Sub Form1_Closing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        ' Disconnect from the RMC, if it is currently connected.
        If RMC IsNot Nothing Then
            If RMC.IsConnected(PingType.DoNotPing) Then
                RMC.Disconnect()
            End If
        End If

        ' Save the quantities from the quantity table to settings, so that
        ' they can be loaded back in when the program is next loaded.
        Dim dt As DataTable = New Data.DataTable("PlotItems")
        dt = CType(Me.dgPlotItems.DataSource, DataTable)
        Dim strColl As New System.Collections.Specialized.StringCollection()
        For Each row As DataRow In dt.Rows
            strColl.Add(row(0).ToString)
            strColl.Add(row(1).ToString)
        Next
        My.Settings.PlotDataItems = strColl
        My.Settings.Save()
    End Sub

    Private Sub connectButton_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ConnectButton.Click
        Try
            Connect()
        Catch ex As ConnectionNotMadeException
            MessageBox.Show("Unable to connect. " + ex.Message)
            UIDisconnect()
        End Try
    End Sub

    Private Sub disconnectButton_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles DisconnectButton.Click
        RMC.Disconnect()
        Disconnect()
    End Sub

    ''' <summary>
    ''' Run when the user attempts to connect to an RMC by clicking the "Connect" button.
    ''' </summary>
    Private Sub Connect()
        ' Create an RMCLink object.
        ' "rmc70.deltamotion.com" is the DNS name for an RMC70 provided by Delta 
        ' that is always available through the web. 
        ' Typically, an IP address such as "192.168.0.161" would be used.
        Dim rmcType As DeviceType
        Select Case RmcTypeComboBox.Text
            Case "RMC70" : rmcType = DeviceType.RMC70
            Case "RMC150" : rmcType = DeviceType.RMC150
            Case "RMC200" : rmcType = DeviceType.RMC200
        End Select

        UIAttemptConnect()

        RMC = RMCLink.CreateEthernetLink(rmcType, IPAddrTextBox.Text)
        RMC.Connect()

        ' We have successfully connected, so update which buttons are enabled.
        UIConnect()

    End Sub

    ''' <summary>
    ''' Run when the user attempts to disconnect from an RMC by clicking the "Disconnect" button,
    ''' OR an attempt to communicate with the RMC fails.
    ''' </summary>
    Private Sub Disconnect()
        ' We have successfully disconnected, so update which buttons are enabled.
        UIDisconnect()

        ' Stop Timer
        AutoPlotUploadTimer.Stop()
    End Sub

    ''' <summary>
    ''' Updates the UI to reflect that the program is connected to an RMC.
    ''' </summary>
    Private Sub UIConnect()
        DisconnectButton.Enabled = True
        ConnectButton.Enabled = False
        RmcTypeComboBox.Enabled = False
        IPAddrTextBox.Enabled = False
        AutoPlotUploadButton.Enabled = True
        ConnectionStatusLabel.Text = "CONNECTED"
        ConnectionStatusLabel.ForeColor = Color.Green
        ConnectionTimer.Enabled = True
        SinglePlotUploadButton.Enabled = True
    End Sub

    ''' <summary>
    ''' Updates the UI to reflect that the program is attempting to connect
    ''' to an RMC.
    ''' </summary>
    Private Sub UIAttemptConnect()
        ConnectionStatusLabel.Text = "CONNECTING..."
        ConnectionStatusLabel.ForeColor = Color.Blue
    End Sub

    ''' <summary>
    ''' Updates the UI to reflect that the program is not connected to an
    ''' RMC.
    ''' </summary>
    Private Sub UIDisconnect()
        ConnectButton.Enabled = True
        DisconnectButton.Enabled = False
        RmcTypeComboBox.Enabled = True
        IPAddrTextBox.Enabled = True

        Dim Gr As Drawing.Graphics
        Gr = PlotDisplay.CreateGraphics
        Gr.Clear(Color.White)
        ConnectionStatusLabel.Text = "DISCONNECTED"
        ConnectionStatusLabel.ForeColor = Color.Red
        PlotScanRadioButton.Checked = False
        AutoPlotUploadButton.Enabled = False
        AutoPlotUploadStopButton.Enabled = False
        SinglePlotUploadButton.Enabled = False
    End Sub

    ''' <summary>
    ''' When auto plot upload is enabled, this function runs several times
    ''' a second to check if the plot in the RMC has been updated (CheckForPlot()).
    ''' 
    ''' If it has been updated, the plot data is downloaded and displayed. Additionally,
    ''' if the "Save Plot to File" checkbox is checked, the plot will also be saved to disk.
    ''' </summary>
    Private Sub AutoPlotUploadTimer_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles AutoPlotUploadTimer.Tick
        ' Check if there is a new plot to upload.
        CheckForPlot()

        ' Toggle the radio button on and off repeatedly to indicate that the program is actively
        ' listening to the RMC.
        If PlotScanRadioButton.Checked Then
            PlotScanRadioButton.Checked = False
        Else
            PlotScanRadioButton.Checked = True
        End If
    End Sub

    ''' <summary>
    ''' Checks the current status of the user-selected plot, to see if there is new plot data
    ''' to upload. If so, the new plot data is uploaded, displayed onscreen, and saved to a plot
    ''' file (if the check box is checked).
    ''' </summary>
    Private Sub CheckForPlot()
        Dim plotState As Integer
        Dim plotID As Integer
        Dim plotConfig As Integer

        ' Only check for plot if window isn't minimized.
        If WindowState <> FormWindowState.Minimized Then
            Try
                ' Confirm that we are still connected to the RMC, before
                ' attempting to retrieve data from it.
                If RMC.IsConnected(PingType.Ping) Then
                    ' Convert the plot configuration number string to an actual integer.
                    plotConfig = Val(PlotConfigComboBox.Text)

                    ' Get the plot ID and state for the given configuration number.
                    GetPlotIDandState(plotConfig, plotID, plotState)

                    ' Different behavior based on plot state:
                    ' 2 = plot is complete; read and display it.
                    ' 1 = plot is currently capturing; display a waiting message.
                    ' 0 = plot has not yet been triggered; wait for it to be triggered.
                    If (plotState = 2) And Not (plotID = LastPlotID) Then
                        ReadAndDisplayPlot(plotID)
                        PlotStatusTextLabel.Text = "Loaded"
                    ElseIf plotState = 1 Then
                        PlotStatusTextLabel.Text = "Loading..."
                    ElseIf plotState = 0 Or plotID = LastPlotID Then
                        PlotStatusTextLabel.Text = "Waiting for new plot..."
                    End If
                End If
            Catch ex As ReadWriteFailedException
                MessageBox.Show("Unable to read data. " + ex.Message)
                Disconnect()
            End Try
        End If
    End Sub

    ''' <summary>
    ''' Run when the user clicks the "Auto Plot Upload" button.
    ''' </summary>
    Private Sub btnAutoPlotUpload_Click(sender As Object, e As EventArgs) Handles AutoPlotUploadButton.Click
        AutoPlotUploadButton.Enabled = False
        AutoPlotUploadStopButton.Enabled = True
        SinglePlotUploadButton.Enabled = False
        AutoPlotUploadTimer.Start()
    End Sub

    ''' <summary>
    ''' Run when the user clicks the "Stop Auto Mode" button.
    ''' </summary>
    Private Sub btnStopAutoMode_Click(sender As Object, e As EventArgs) Handles AutoPlotUploadStopButton.Click
        StopAutoMode()
    End Sub

    ''' <summary>
    ''' Stops the auto plot upload mode, including updating the UI accordingly.
    ''' </summary>
    Private Sub StopAutoMode()
        AutoPlotUploadStopButton.Enabled = False
        AutoPlotUploadButton.Enabled = True
        PlotScanRadioButton.Checked = False
        SinglePlotUploadButton.Enabled = True
        AutoPlotUploadTimer.Stop()
    End Sub

    ''' <summary>
    ''' Gets information regarding the plot that will be necessary for reading it.
    ''' </summary>
    ''' <param name="plotConfig">The configuration number of the plot to get information for.</param>
    ''' <param name="dataItems">The number of quantities recorded by the plot.</param>
    ''' <param name="plotSamples">The number of samples the plot intends to capture.</param>
    ''' <param name="capturedSamples">The number of samples captured by the plot. This may be less than plotSamples if not all samples could be captured.</param>
    ''' <param name="plotPeriod">The number of seconds the plot was captured over.</param>
    ''' <param name="regAddr">A list of addresses in IEC format recorded by the plot.</param>
    Private Sub GetPlotConfig(plotConfig As Integer, ByRef dataItems As Integer, ByRef plotSamples As Integer,
                          ByRef capturedSamples As Integer, ByRef plotPeriod As Single, ByRef regAddr() As String)
        Dim i

        ' Array for storing plot configuration information from the RMC.
        Dim PlotData() As Single = New Single(32) {}

        ' Used for the RMC200 specifically.
        Dim inPlotData() As Integer = New Integer(32) {}

        ' Read plot data depending on the RMC type.
        Select Case RmcTypeComboBox.Text
            Case "RMC70"
                RMC.ReadFFile(32 + plotConfig, 1, PlotData, 0, 32)
                plotSamples = PlotData(0)
                plotPeriod = PlotData(1)
            Case "RMC150"
                RMC.ReadFFile(96 + plotConfig, 1, PlotData, 0, 32)
                plotSamples = PlotData(0)
                plotPeriod = PlotData(1)
            Case "RMC200"
                RMC.ReadLFile(512 + plotConfig, 1, inPlotData, 0, 32)
                RMC.ReadFFile(512 + plotConfig, 2, PlotData, 0, 1)
                plotPeriod = PlotData(0)
                plotSamples = inPlotData(0)

                For i = 0 To 31
                    PlotData(i) = CSng(inPlotData(i))
                Next i
        End Select

        capturedSamples = PlotData(9)

        ' Extract the number of data items and register addresses from the plot configuration.
        dataItems = 0
        i = 0
        While i < 16 And PlotData(15 + i) <> 0
            dataItems += 1
            regAddr(i) = "%MD" + CStr(Int(PlotData(15 + i) / 4096)) + "." +
                             CStr(PlotData(15 + i) Mod 4096)
            i += 1
        End While
        dataItems = i
    End Sub

    ''' <summary>
    ''' Gets the plot ID and state.
    ''' 
    ''' Each time a new plot is generated, the ID changes. By comparing two IDs for the same
    ''' configuration number, we can detect when a new plot has been created.
    ''' </summary>
    ''' <param name="plotConfig">The plot configuration number to get the ID and state for.</param>
    ''' <param name="plotID">The ID of the plot with the provided plot configuration number.</param>
    ''' <param name="plotState">The state (not triggered, currently capturing, or complete) of the plot with the provided plot configuration number.</param>
    Private Sub GetPlotIDandState(plotConfig As Integer, ByRef plotID As Integer, ByRef plotState As Integer)
        Dim PlotData() As Single = New Single(1) {}
        Dim inPlotData() As Integer = New Integer(1) {}

        Select Case RmcTypeComboBox.Text
            Case "RMC70"
                RMC.ReadFFile(32 + plotConfig, 8, PlotData, 0, 2)
                plotID = PlotData(0)
                plotState = PlotData(1)
            Case "RMC150"
                RMC.ReadFFile(96 + plotConfig, 8, PlotData, 0, 2)
                plotID = PlotData(0)
                plotState = PlotData(1)
            Case "RMC200"
                RMC.ReadLFile(512 + plotConfig, 8, inPlotData, 0, 2)
                plotID = inPlotData(0)
                plotState = inPlotData(1)
        End Select
    End Sub

    ''' <summary>
    ''' Run several times a second to determine whether or not we are still connected
    ''' to the RMC. If we are are no longer connected, then the UI is updated to reflect that.
    ''' </summary>
    Private Sub ConnectionTimer_Tick(sender As Object, e As EventArgs) Handles ConnectionTimer.Tick
        Try
            If Not RMC.IsConnected(PingType.Ping) Then
                Disconnect()
            End If
        Catch ex As Exception
            Disconnect()
        End Try
    End Sub

    ''' <summary>
    ''' Reads the requested plot configuration using the Dynamic Plot Upload Area,
    ''' which allows for reading long plots. The RMCTools Help documentation describes
    ''' how to use the Dynamic Plot Upload area in the topic "Reading Plots with a Host Controller".
    ''' </summary>
    ''' <param name="plotConfig">The configuration number of the plot to read.</param>
    ''' <param name="dataItems">The number of data sets in the plot to read.</param>
    ''' <param name="capturedSamples">The number of plot samples to read.</param>
    ''' <returns></returns>
    Private Function ReadPlot(plotConfig, dataItems, capturedSamples) As Single(,)
        Dim addrFilePlots, readLength As Integer
        Dim WriteData() As Integer = New Integer(1) {}
        Dim plotInfo() As Integer = New Integer(4) {}
        Dim i, j, k As Integer

        Select Case RmcTypeComboBox.Text
            Case "RMC70" : addrFilePlots = 40
            Case "RMC150" : addrFilePlots = 104
            Case "RMC200" : addrFilePlots = 576
            Case Else
                addrFilePlots = 0
        End Select

        ' Write 1 to the Upload Mode/Status Register and write the number of requested read samples.
        WriteData(0) = 1

        ' The number of samples read each time is 200 x plotItems.
        WriteData(1) = 200


        RMC.WriteLFile(addrFilePlots + plotConfig, 0, WriteData, 0, 2)

        'Read the plot data 200 samples at a time.
        'Each read length will need to be 200 x plotItems + 5, but the 5 will be done with a ReadLFile.
        readLength = 200 * dataItems

        'After each read, extract the data and put in plotSet array
        '(data item, sample)
        Dim plotSet(,) As Single = New Single(capturedSamples - 1, dataItems - 1) {}
        Dim PlotDataArray() As Single = New Single(readLength) {}

        ' Keep reading plot data and appending it to the plotSet array.
        ' (The conversion to int is for truncating.)
        For i = 0 To Int(capturedSamples / 200) - 1

            ' Read the data from the RMC.
            RMC.ReadLFile(addrFilePlots + plotConfig, 0, plotInfo, 0, 5)
            RMC.ReadFFile(addrFilePlots + plotConfig, 5, PlotDataArray, 0, readLength)

            ' Extract the plotSet data.
            For j = 0 To dataItems - 1
                For k = 0 To 199
                    plotSet(i * 200 + k, j) = PlotDataArray(j * 200 + k)
                Next k
            Next j
        Next i

        ' Do the last read.
        If (capturedSamples Mod 200) <> 0 Then
            ' Write the number of requested read samples for the last read.
            WriteData(0) = capturedSamples Mod 200
            RMC.WriteLFile(addrFilePlots + plotConfig, 1, WriteData, 0, 1)

            ' Read the data from the RMC.
            RMC.ReadLFile(addrFilePlots + plotConfig, 0, plotInfo, 0, 5)
            RMC.ReadFFile(addrFilePlots + plotConfig, 5, PlotDataArray, 0, (capturedSamples Mod 200) * dataItems)

            ' Extract the plotSet data.
            For j = 0 To dataItems - 1
                For k = 0 To (capturedSamples Mod 200) - 1
                    plotSet(i * 200 + k, j) = PlotDataArray(j * (capturedSamples Mod 200) + k)
                Next
            Next
        End If

        ReadPlot = plotSet
    End Function

    ''' <summary>
    ''' Run when the user requests a single plot upload by clicking the "Single Plot Upload" button.
    ''' </summary>
    Private Sub ButtonSingleUpload_Click(sender As Object, e As EventArgs) Handles SinglePlotUploadButton.Click
        CheckForPlot()
    End Sub

    ''' <summary>
    ''' Reads the requested plot, and displays it on-screen.
    ''' 
    ''' First, it reads the plot configuration information to find out
    ''' how many quantities and samples are in the requested plot. Then,
    ''' it reads the plot data and displays it in a chart.
    ''' </summary>
    ''' <param name="PlotID">The ID of the plot to read.</param>
    Private Sub ReadAndDisplayPlot(PlotID As Integer)
        Dim plotSamples, capturedSamples, plotItems, plotConfig As Integer
        Dim plotPeriod As Single
        Dim i, j As Integer
        Dim regAddr(15) As String

        plotConfig = CInt(PlotConfigComboBox.Text)

        ' Get the plotSamples, plotPeriod, plotItems, and register addresses.
        GetPlotConfig(plotConfig, plotItems, plotSamples, capturedSamples, plotPeriod, regAddr)

        ' There should be as many quantities defined in the "Plot Quantities" table
        ' as there are quantities in the plot on the RMC. If these two values are not equal,
        ' we have to error out and wait for the user to fix it.
        If Not plotItems = (dgPlotItems.Rows.Count - 1) Then
            StopAutoMode()
            MsgBox("Error: The plot items list must have the same number items as the plot in the RMC.  RMC Plot items = " + Str(plotItems) + " Display Plot Items = " + Str(dgPlotItems.Rows.Count - 1))
            Exit Sub
        End If

        ' We are successfully reading this plot. As a result, we want to update our stored
        ' last plot ID to be the current plot ID; that way, we will know we're not wasting
        ' time or energy redownloading the same plot data.
        LastPlotID = PlotID

        ' Read the plot data into an array.
        Dim plotData(,) = ReadPlot(plotConfig, plotItems, capturedSamples)

        ' Make the x axis data (in seconds).
        Dim plotTIme() As Single = New Single(plotSamples - 1) {}
        For i = 0 To plotSamples - 1
            plotTime(i) = plotPeriod * i
        Next i

        ' Array of quantity names corresponding to values in plotData.
        Dim itemNames() As String = New String(plotItems - 1) {}

        ' Array of quantity units corresponding to values in plotData.
        Dim Units() As String = New String(plotItems - 1) {}

        PlotDisplay.Series.Clear()
        For i = 0 To plotItems - 1
            ' Set this quantity's name (e.g. "Actual Position").
            ' If nothing is specified, it will be an empty string.
            itemNames(i) = dgPlotItems.Rows(i).Cells(0).Value.ToString()

            ' Set this quantity's units (e.g. "pu").
            ' If nothing is specified, it will be an empty string.
            Units(i) = dgPlotItems.Rows(i).Cells(1).Value.ToString()

            ' When displaying on the plot, we want the format
            ' "Quantity Name (units)", or just "Quantity Name" if no
            ' units are specified (i.e. Units(i) is an empty string).
            Dim displayString = itemNames(i)
            If Units(i) <> "" Then
                displayString += " (" + Units(i) + ")"
            End If

            ' Add this quantity to the chart as a series.
            PlotDisplay.Series.Add(displayString)
            PlotDisplay.Series(i).Points.Clear()
            PlotDisplay.Series(i).ChartType = DataVisualization.Charting.SeriesChartType.Line
        Next

        ' Insert each data XY pair into the chart.
        For j = 0 To plotItems - 1
            For i = 0 To UBound(plotData, 1)
                PlotDisplay.Series(j).Points.AddXY(plotTime(i), plotData(i, j))
            Next i
        Next j

        ' Display the chart.
        PlotDisplay.Update()

        ' If the user has requested it by checking the checkbox, then save
        ' the plot data, either as an .rmcplots or .csv file.
        If SaveToFileCheckbox.Checked Then
            If ComboBoxFileDataType.SelectedItem.ToString = "RMC Plot" Then
                PlotCreateXMLFile(plotData, regAddr, itemNames, Units, plotPeriod)
            Else
                PlotCreateCSVfile(plotData, itemNames)
            End If
        End If

    End Sub

    ''' <summary>
    ''' Run when the user attempts to save their current configuration by clicking the
    ''' "Save Settings" button.
    ''' </summary>
    Private Sub ButtonSaveToFile_Click(sender As Object, e As EventArgs) Handles ButtonSaveToFile.Click
        ' Configure the "Save Settings" dialog box.
        Dim fd As SaveFileDialog = New SaveFileDialog()
        fd.Title = "Save Settings"
        fd.InitialDirectory = "C:\"
        fd.Filter = "All files (*.*)|*.*|ini files (*.ini)|*.ini"
        fd.FilterIndex = 2
        fd.RestoreDirectory = True

        ' Show the "Save Settings" dialog box, and wait for a response.
        ' If the user selected a destination, then write the configuration
        ' information to it.
        If fd.ShowDialog() = DialogResult.OK Then
            Using sw As New System.IO.StreamWriter(fd.FileName, False)
                sw.WriteLine(RmcTypeComboBox.Text)
                sw.WriteLine(IPAddrTextBox.Text)
                sw.WriteLine(ComboBoxFileDataType.Text)
                sw.WriteLine(PlotPathTextBox.Text)
                sw.WriteLine(PlotNameTextBox.Text)
                sw.WriteLine(PlotConfigComboBox.Text)
                sw.WriteLine(SaveToFileCheckbox.Checked)
                Dim dt As DataTable = New Data.DataTable("PlotItems")
                dt = CType(Me.dgPlotItems.DataSource, DataTable)
                For Each row As DataRow In dt.Rows
                    sw.WriteLine(row(0).ToString)
                    sw.WriteLine(row(1).ToString)
                Next
            End Using
        End If


    End Sub

    ''' <summary>
    ''' Run when the user attempts to load a configuration by clicking the "Load Settings" button.
    ''' </summary>
    Private Sub ButtonReadFromFile_Click(sender As Object, e As EventArgs) Handles ButtonReadFromFile.Click
        ' Configure the "Load Settings" dialog box.
        Dim fd As OpenFileDialog = New OpenFileDialog()
        fd.Title = "Load Settings"
        fd.InitialDirectory = "C:\"
        fd.Filter = "All files (*.*)|*.*|ini files (*.ini)|*.ini"
        fd.FilterIndex = 2
        fd.RestoreDirectory = True

        ' Show the "Load Settings" dialog box, and wait for a response.
        ' If the user selected a destination, then load the configuration
        ' from it.
        If fd.ShowDialog() = DialogResult.OK Then
            Using sr As New System.IO.StreamReader(fd.FileName, False)
                RmcTypeComboBox.Text = sr.ReadLine
                IPAddrTextBox.Text = sr.ReadLine
                ComboBoxFileDataType.Text = sr.ReadLine
                PlotPathTextBox.Text = sr.ReadLine
                PlotNameTextBox.Text = sr.ReadLine
                PlotConfigComboBox.Text = sr.ReadLine
                SaveToFileCheckbox.Checked = sr.ReadLine
                Dim Col0 As String = ""
                Dim Col1 As String = ""
                Dim dt As DataTable = New Data.DataTable("ParentTable")
                dt.Columns.Add("Item Name", GetType(String))
                dt.Columns.Add("Units", GetType(String))
                Me.dgPlotItems.DataSource = dt
                Col0 = sr.ReadLine
                Col1 = sr.ReadLine
                Do While Col0 IsNot Nothing
                    dt.Rows.Add(New String() {Col0, Col1})
                    Col0 = sr.ReadLine
                    Col1 = sr.ReadLine
                Loop
            End Using
        End If
    End Sub

    ''' <summary>
    ''' Run when the user attempts to select a destination folder for uploaded plots.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    Private Sub PlotPathBrowseButton_Click(sender As Object, e As EventArgs) Handles PlotPathBrowseButton.Click
        ' Configure the folder selection dialog.
        Dim fd As FolderBrowserDialog = New FolderBrowserDialog()
        fd.SelectedPath = PlotPathTextBox.Text

        ' If a folder was selected, then update the PlotPathTextBox to hold that path.
        If fd.ShowDialog() = DialogResult.OK Then
            PlotPathTextBox.Text = fd.SelectedPath
        End If
    End Sub

    ''' <summary>
    ''' Gets the loop time of the controller
    ''' </summary>
    ''' <returns>A single representing the controller's loop time</returns>
    Private Function GetLoopTime() As Single
        Dim readData() As Single = New Single(1) {}

        Select Case RmcTypeComboBox.Text
            Case "RMC70" : RMC.ReadFFile(20, 0, readData, 0, 1)
            Case "RMC150" : RMC.ReadFFile(44, 0, readData, 0, 1)
            Case "RMC200" : RMC.ReadFFile(18, 0, readData, 0, 1)
        End Select

        GetLoopTime = readData(0)
        Exit Function
    End Function

    ''' <summary>
    ''' Given plot data, constructs an .rmcplots file and saves it to the folder
    ''' specified by the user.
    ''' </summary>
    ''' <param name="plotSet">The plot data.</param>
    ''' <param name="regAddr">An array consisting of the register addresses as strings.</param>
    ''' <param name="itemNames">An array consisting of the quantity names.</param>
    ''' <param name="units">An array consisting of the quantity units.</param>
    ''' <param name="plotPeriod">The plot period.</param>
    Private Sub PlotCreateXMLFile(plotSet, regAddr, itemNames, units, plotPeriod)
        Dim i, j As Integer
        Dim vbTab2
        Dim vbTab3
        Dim fileName As String
        Dim xmlData As String

        fileName = CreateFilename() + ".rmcplots"

        ' Save tab sizes for indentation in plot file for later use.
        vbTab2 = vbTab + vbTab
        vbTab3 = vbTab + vbTab + vbTab

        ' Create the object that will write the plot data to disk.
        Dim objWriter As New IO.StreamWriter(fileName)

        ' Pen colors for different quantities.
        Dim penColors(16)
        penColors(0) = "16#E0E000"
        penColors(1) = "16#0000FF"
        penColors(2) = "16#FF00FF"
        penColors(3) = "16#FF0000"
        penColors(4) = "16#00E000"
        penColors(5) = "16#E0E000"
        penColors(6) = "16#00E0E0"
        penColors(7) = "16#C0C0C0"
        penColors(8) = "16#0066FF"
        penColors(9) = "16#CC99FF"
        penColors(10) = "16#CC99FF"
        penColors(11) = "16#663399"
        penColors(12) = "16#FF99CC"
        penColors(13) = "16#003399"
        penColors(14) = "16#00CC99"
        penColors(15) = "16#808000"

        ' Create beginning of XML data
        xmlData = "<?xml version='1.0' encoding='utf-8'?>" + vbCrLf
        xmlData += "<PlotFolderRoot" + vbCrLf
        xmlData += vbTab + "Name=""Plot1""" + vbCrLf
        xmlData += vbTab + "Comment=""""" + vbCrLf
        xmlData += vbTab + "PlotFileVersion = ""1.1"">" + vbCrLf
        xmlData += vbTab + "<Plot" + vbCrLf
        xmlData += vbTab2 + "Name=""Controller1:NoName:1""" + vbCrLf
        xmlData += vbTab2 + "Comment = """"" + vbCrLf
        xmlData += vbTab2 + "SamplePeriodS=" + """" + CStr(plotPeriod) + """" + vbCrLf
        xmlData += vbTab2 + "CapturedDT=""9999-01-01 0 01:01:01.001""" + vbCrLf
        xmlData += vbTab2 + "LoopsPerInt=""1""" + vbCrLf
        xmlData += vbTab2 + "Items=""" + CStr(UBound(plotSet, 2) + 1) + """" + vbCrLf
        xmlData += vbTab2 + "FirstIntEver=""0"">" + vbCrLf
        xmlData += vbTab2 + "<Controller" + vbCrLf
        xmlData += vbTab3 + "ControllerType =" + """" + RmcTypeComboBox.Text + """" + "/>" + vbCrLf

        ' Create Item elements.
        Dim itemLabels(,) As String = New String(UBound(plotSet, 2), 3) {}
        For i = 0 To UBound(plotSet, 2)
            ' Add item opening tag.
            itemLabels(i, 1) = "<Item"

            ' Add address of item.
            itemLabels(i, 2) = "Address=""" & CStr(regAddr(i)) & """"

            ' Add type attribute of item.
            itemLabels(i, 3) = "Type=""REAL""/>"
        Next

        ' Add item elements to plot file buffer.
        For i = 0 To UBound(plotSet, 2)
            xmlData += vbTab2 + itemLabels(i, 1) + vbCrLf
            xmlData += vbTab3 + itemLabels(i, 2) + vbCrLf
            xmlData += vbTab3 + itemLabels(i, 3) + vbCrLf
        Next i

        ' Create plot quantity labels.
        Dim quantityLabels(,) As String = New String(UBound(plotSet, 1), 10) {}
        For i = 0 To UBound(plotSet, 2)
            ' Add quantity opening tag.
            quantityLabels(i, 0) = "<Quantity"

            ' Add name attribute for quantity.
            quantityLabels(i, 1) = "Name=" & """" & CStr(itemNames(i)) & """"

            ' Add units attribute for quantity.
            quantityLabels(i, 2) = "Units=" & """" & CStr(units(i)) & """"

            ' Add formula (i.e. register address) attribute for quantity.
            quantityLabels(i, 3) = "Formula=" & """" & CStr(regAddr(i)) & """"

            ' Add several static attributes for quantity.
            quantityLabels(i, 4) = "Filter=""False"""
            quantityLabels(i, 5) = "Type=""REAL"""
            quantityLabels(i, 6) = "DecPlaces=""3"""
            quantityLabels(i, 7) = "Graph=""True"""

            ' Add pen attribute (color of the quantity when drawn in RMCTools)
            ' for quantity.
            quantityLabels(i, 8) = "Pen=" & """" & CStr(penColors(i)) & """"

            ' Add AutoScale and AutoGroup attributes for quantity.
            quantityLabels(i, 9) = "AutoScale=""True"""
            quantityLabels(i, 10) = "AutoGroup=""True""/>"
        Next i

        ' Add quanitity labels to plot file buffer.
        For i = 0 To UBound(plotSet, 2)
            xmlData += vbTab2 & quantityLabels(i, 0) + vbCrLf
            For j = 1 To 10
                xmlData += vbTab3 & quantityLabels(i, j) + vbCrLf
            Next j
        Next i

        ' Write items and quantities to disk.
        objWriter.Write(xmlData)

        ' Reset the buffer string to be empty, so we can add
        ' the sample values to it.
        xmlData = vbNullString

        ' Create the plot data array.
        Dim plotValues() As String = New String(UBound(plotSet, 1)) {}
        For i = 0 To UBound(plotSet, 1)
            ' Start the sample and include its interval.
            xmlData += vbTab2 + "<Sample Int=""" + CStr(i) + """"

            ' For each item, include the item's value at the time of the sample.
            For j = 0 To UBound(plotSet, 2)
                xmlData += " Item" + CStr(j) + "=""" + CStr(plotSet(i, j)) + """"
            Next j

            ' Finally, close the sample tag.
            xmlData += " />" + vbCrLf

            ' Write this sample to disk, and then empty the string buffer again.
            objWriter.Write(xmlData)
            xmlData = vbNullString
        Next i

        ' Add closing tags.
        xmlData += vbTab + "</Plot>" + vbCrLf
        xmlData += "</PlotFolderRoot>"

        ' Write the closing tags to disk, then close the file.
        objWriter.Write(xmlData)
        objWriter.Close()
    End Sub

    ''' <summary>
    ''' Given plot data, constructs a .csv file and saves it to the folder specified
    ''' by the user.
    ''' </summary>
    ''' <param name="plotSet">The plot data.</param>
    ''' <param name="itemNames">An array consisting of the quantity names.</param>
    Private Sub PlotCreateCSVfile(plotSet, itemNames)
        Dim i, j As Integer
        Dim fileName As String
        Dim csvData As String = ""

        fileName = CreateFilename() + ".csv"

        ' Create the object that will write the plot data to disk.
        Dim objWriter As New IO.StreamWriter(fileName)

        ' Create plot data array
        Dim plotValues() As String = New String(UBound(plotSet, 1)) {}

        ' Write the quantity names as the first row in the .csv file.
        For i = 0 To UBound(plotSet, 2)
            If i > 0 Then csvData += ","
            csvData += CStr(itemNames(i))
        Next

        ' Add a return to the output buffer.
        csvData += vbCrLf

        ' Write the output buffer to disk.
        objWriter.Write(csvData)

        ' Clear the output buffer.
        csvData = vbNullString

        ' Add each sample's values to the output buffer.
        For j = 0 To UBound(plotSet, 1)
            For i = 0 To UBound(plotSet, 2)
                If i > 0 Then csvData += ","
                csvData += CStr(plotSet(j, i))
            Next
            csvData += vbCrLf

            ' We write out each line to the file to prevent the string from getting too big.
            ' When the string gets big, performance gets terrible.
            objWriter.Write(csvData)
            csvData = vbNullString
        Next j

        ' Write the closing tags to disk, then close the file.
        objWriter.Write(csvData)
        objWriter.Close()

    End Sub

    ''' <summary>
    ''' Create a plot filename as a combination of the user-specified prefix string
    ''' and the current date and time
    ''' </summary>
    ''' <returns></returns>
    Private Function CreateFilename()
        Dim d As Date = Now
        CreateFilename = $"{PlotPathTextBox.Text}\{PlotNameTextBox.Text}_{d.Year}_{d.Month}_{d.Day} {d.Hour} {d.Minute} {d.Second} {d.Millisecond}"
    End Function
End Class
